##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from collections import Counter
from scipy import optimize
import numpy as np
from grove.pyvqe.vqe import VQE
import pyquil.quil as pq
from pyquil.gates import H
from pyquil.paulis import exponential_map, PauliSum
from functools import reduce


class QAOA(object):
    def __init__(self, qvm, qubits, steps=1, init_betas=None,
                 init_gammas=None, cost_ham=None,
                 ref_ham=None, driver_ref=None,
                 minimizer=None, minimizer_args=None,
                 minimizer_kwargs=None, rand_seed=None,
                 vqe_options=None, store_basis=False):
        """
        QAOA object.

        Contains all information for running the QAOA algorthm to find the
        ground state of the list of cost clauses.

        N.B. This only works if all the terms in the cost Hamiltonian commute with each other.

        :param qvm: (Connection) The qvm connection to use for the algorithm.
        :param qubits: (list of ints) The number of qubits to use for the algorithm.
        :param steps: (int) The number of mixing and cost function steps to use.
                      Default=1.
        :param init_betas: (list) Initial values for the beta parameters on the
                           mixing terms. Default=None.
        :param init_gammas: (list) Initial values for the gamma parameters on the
                            cost function. Default=None.
        :param cost_ham: list of clauses in the cost function. Must be
                    PauliSum objects
        :param ref_ham: list of clauses in the mixer function. Must be
                    PauliSum objects
        :param driver_ref: (pyQuil.quil.Program()) object to define state prep
                           for the starting state of the QAOA algorithm.
                           Defaults to tensor product of \|+> states.
        :param rand_seed: integer random seed for initial betas and gammas
                          guess.
        :param minimizer: (Optional) Minimization function to pass to the
                          Variational-Quantum-Eigensolver method
        :param minimizer_kwargs: (Optional) (dict) of optional arguments to pass to
                                 the minimizer.  Default={}.
        :param minimizer_args: (Optional) (list) of additional arguments to pass to the
                               minimizer. Default=[].
        :param minimizer_args: (Optional) (list) of additional arguments to pass to the
                               minimizer. Default=[].
        :param vqe_options: (optinal) arguents for VQE run.
        :param store_basis: (optional) boolean flag for storing basis states.
                            Default=False.
        """

        # Seed the random number generator, if a seed is provided.
        if rand_seed is not None:
            np.random.seed(rand_seed)

        # Set attributes values, considering their defaults
        self.qvm = qvm
        self.steps = steps
        self.qubits = qubits
        self.nstates = 2 ** len(qubits)

        self.cost_ham = cost_ham or []
        self.ref_ham = ref_ham or []

        self.minimizer = minimizer or optimize.minimize
        self.minimizer_args = minimizer_args or []
        self.minimizer_kwargs = minimizer_kwargs or {
            'method': 'Nelder-Mead',
            'options': {
                'disp': True,
                'ftol': 1.0e-2,
                'xtol': 1.0e-2
            }
        }

        self.betas = init_betas or np.random.uniform(0, np.pi, self.steps)[::-1]
        self.gammas = init_gammas or np.random.uniform(0, 2*np.pi, self.steps)
        self.vqe_options = vqe_options or {}

        self.ref_state_prep = (
            driver_ref or
            pq.Program([H(i) for i in self.qubits])
        )

        if store_basis:
            self.states = [
                np.binary_repr(i, width=len(self.qubits))
                for i in range(self.nstates)
            ]

        # Check argument types
        if not isinstance(self.cost_ham, (list, tuple)):
            raise TypeError("cost_ham must be a list of PauliSum objects.")
        if not all([isinstance(x, PauliSum) for x in self.cost_ham]):
            raise TypeError("cost_ham must be a list of PauliSum objects")

        if not isinstance(self.ref_ham, (list, tuple)):
            raise TypeError("ref_ham must be a list of PauliSum objects")
        if not all([isinstance(x, PauliSum) for x in self.ref_ham]):
            raise TypeError("ref_ham must be a list of PauliSum objects")

        if not isinstance(self.ref_state_prep, pq.Program):
            raise TypeError("Please provide a pyQuil Program object "
                            "to generate initial state.")

    def get_parameterized_program(self):
        """
        Return a function that accepts parameters and returns a new Quil
        program.

        :returns: a function
        """
        cost_para_programs = []
        driver_para_programs = []

        for idx in range(self.steps):
            cost_list = []
            driver_list = []
            for cost_pauli_sum in self.cost_ham:
                for term in cost_pauli_sum.terms:
                    cost_list.append(exponential_map(term))

            for driver_pauli_sum in self.ref_ham:
                for term in driver_pauli_sum.terms:
                    driver_list.append(exponential_map(term))

            cost_para_programs.append(cost_list)
            driver_para_programs.append(driver_list)

        def psi_ref(params):
            """
            Construct a Quil program for the vector (beta, gamma).

            :param params: array of 2 . p angles, betas first, then gammas
            :return: a pyquil program object
            """
            if len(params) != 2*self.steps:
                raise ValueError("params doesn't match the number of parameters set by `steps`")
            betas = params[:self.steps]
            gammas = params[self.steps:]

            prog = pq.Program()
            prog += self.ref_state_prep
            for idx in range(self.steps):
                for fprog in cost_para_programs[idx]:
                    prog += fprog(gammas[idx])

                for fprog in driver_para_programs[idx]:
                    prog += fprog(betas[idx])

            return prog

        return psi_ref

    def get_angles(self):
        """
        Finds optimal angles with the quantum variational eigensolver method.

        Stored VQE result

        :returns: ([list], [list]) A tuple of the beta angles and the gamma
                  angles for the optimal solution.
        """
        stacked_params = np.hstack((self.betas, self.gammas))
        vqe = VQE(self.minimizer, minimizer_args=self.minimizer_args,
                  minimizer_kwargs=self.minimizer_kwargs)
        cost_ham = reduce(lambda x, y: x + y, self.cost_ham)
        # maximizing the cost function!
        param_prog = self.get_parameterized_program()
        result = vqe.vqe_run(param_prog, cost_ham, stacked_params, qvm=self.qvm,
                             **self.vqe_options)
        self.result = result
        betas = result.x[:self.steps]
        gammas = result.x[self.steps:]
        return betas, gammas

    def probabilities(self, angles):
        """
        Computes the probability of each state given a particular set of angles.

        :param angles: [list] A concatenated list of angles [betas]+[gammas]
        :return: [list] The probabilities of each outcome given those angles.
        """
        if isinstance(angles, list):
            angles = np.array(angles)

        assert angles.shape[0] == 2 * self.steps, "angles must be 2 * steps"

        param_prog = self.get_parameterized_program()
        prog = param_prog(angles)
        wf = self.qvm.wavefunction(prog)
        wf = wf.amplitudes.reshape((-1, 1))
        probs = np.zeros_like(wf)
        for xx in range(2 ** len(self.qubits)):
            probs[xx] = np.conj(wf[xx]) * wf[xx]
        return probs

    def get_string(self, betas, gammas, samples=100):
        """
        Compute the most probable string.

        The method assumes you have passed init_betas and init_gammas with your
        pre-computed angles or you have run the VQE loop to determine the
        angles.  If you have not done this you will be returning the output for
        a random set of angles.

        :param betas: List of beta angles
        :param gammas: List of gamma angles
        :param samples: (int, Optional) number of samples to get back from the
                        QVM.
        :returns: tuple representing the bitstring, Counter object from
                  collections holding all output bitstrings and their frequency.
        """
        if samples <= 0 and not isinstance(samples, int):
            raise ValueError("samples variable must be positive integer")
        param_prog = self.get_parameterized_program()
        stacked_params = np.hstack((betas, gammas))
        sampling_prog = param_prog(stacked_params)

        bitstring_samples = self.qvm.run_and_measure(sampling_prog,
                                                     self.qubits,
                                                     trials=samples)
        bitstring_tuples = list(map(tuple, bitstring_samples))
        freq = Counter(bitstring_tuples)
        most_frequent_bit_string = max(freq, key=lambda x: freq[x])
        return most_frequent_bit_string, freq



##############################################################################
# Copyright 2017-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

"""
Utilities for encapsulating bases and properties of quantum operators and super-operators
as represented by qutip.Qobj()'s.
"""

import itertools
import logging
from itertools import product as cartesian_product

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import tqdm
from matplotlib.colors import LinearSegmentedColormap
from mpl_toolkits.mplot3d import Axes3D
from pyquil.api.errors import QPUError
from pyquil.gates import I, X, H, CZ
from pyquil.quil import Program
from pyquil.quilbase import Pragma

try:
    # Python 2
    from itertools import izip
except ImportError:  # pragma no coverage
    # Python 3
    izip = zip

_log = logging.getLogger(__name__)


_QUTIP_ERROR_LOGGED = False


def import_qutip():
    """
    Try importing the qutip module, log an error if unsuccessful.

    :return: The qutip module if successful or None
    :rtype: Optional[module]
    """
    global _QUTIP_ERROR_LOGGED
    try:
        import qutip
    except ImportError:  # pragma no coverage
        qutip = None
        if not _QUTIP_ERROR_LOGGED:
            _log.error("Could not import qutip. Tomography tools will not function.")
            _QUTIP_ERROR_LOGGED = True
    return qutip


_CVXPY_ERROR_LOGGED = False


def import_cvxpy():
    """
    Try importing the qutip module, log an error if unsuccessful.

    :return: The cvxpy module if successful or None
    :rtype: Optional[module]
    """
    global _CVXPY_ERROR_LOGGED
    try:
        import cvxpy
    except ImportError:  # pragma no coverage
        cvxpy = None
        if not _CVXPY_ERROR_LOGGED:
            _log.error("Could not import cvxpy. Tomography tools will not function.")
            _CVXPY_ERROR_LOGGED = True
    return cvxpy


THREE_COLOR_MAP = ['#48737F', '#FFFFFF', '#D6619E']
rigetti_3_color_cm = LinearSegmentedColormap.from_list("Rigetti", THREE_COLOR_MAP[::-1], N=100)

FIVE_COLOR_MAP = ['#C671A2', '#545253', '#85B5BE', '#ECE9CC', '#C671A2']
rigetti_4_color_cm = LinearSegmentedColormap.from_list("Rigetti", FIVE_COLOR_MAP[::-1], N=100)
EPS = 1e-2
SEED = 137
BELL_STATE_PROGRAM = Program([H(0), H(1), CZ(0, 1), H(1)])
CNOT_PROGRAM = Program([H(1), CZ(0, 1), H(1)])

BAD_1Q_READOUT = np.array([[.9, .15],
                           [.1, .85]])
BAD_2Q_READOUT = np.kron(BAD_1Q_READOUT, BAD_1Q_READOUT)


# constants to provide optional fancy progress bars
NOTEBOOK_MODE = False
TRANGE = tqdm.trange


def notebook_mode(m):
    """
    Configure whether this module should assume that it is being run from a jupyter notebook.
    This sets some global variables related to how progress for long measurement sequences is
    indicated.

    :param bool m: If True, assume to be in notebook.
    :return: None
    :rtype: NoneType
    """
    global NOTEBOOK_MODE
    global TRANGE
    NOTEBOOK_MODE = m
    if NOTEBOOK_MODE:
        TRANGE = tqdm.tnrange
    else:
        TRANGE = tqdm.trange


def to_density_matrix(state):
    """
    Convert a Hilbert space vector to a density matrix.

    :param qt.basis state: The state to convert into a density matrix.
    :return: The density operator corresponding to state.
    :rtype: qutip.qobj.Qobj
    """
    return state * state.dag()


def sample_outcomes(probs, n):
    """
    For a discrete probability distribution ``probs`` with outcomes 0, 1, ..., k-1 draw ``n``
    random samples.

    :param list probs: A list of probabilities.
    :param Number n: The number of random samples to draw.
    :return: An array of samples drawn from distribution probs over 0, ..., len(probs) - 1
    :rtype: numpy.ndarray
    """
    dist = np.cumsum(probs)
    rs = np.random.rand(n)
    return np.array([(np.where(r < dist)[0][0]) for r in rs])


def basis_state_preps(*qubits):
    """
    Generate a sequence of programs that prepares the measurement
    basis states of some set of qubits in the order such that the qubit
    with highest index is iterated over the most quickly:
    E.g., for ``qubits=(0, 1)``, it returns the circuits::

        I_0 I_1
        I_0 X_1
        X_0 I_1
        X_0 X_1

    :param list qubits: Each qubit to include in the basis state preparation.
    :return: Yields programs for each basis state preparation.
    :rtype: Program
    """
    for prep in cartesian_product([I, X], repeat=len(qubits)):
        basis_prep = Program(Pragma("PRESERVE_BLOCK"))
        for gate, qubit in zip(prep, qubits):
            basis_prep.inst(gate(qubit))
        basis_prep.inst(Pragma("END_PRESERVE_BLOCK"))
        yield basis_prep


def basis_labels(n):
    """
    Generate a list of basis labels for `n` qubits, ordered from least to greatest, in big-endian
     format:

        ['00..00', '00..01', ..., '11..11']

    :param n:
    :return: A list of strings of length n that enumerate the n-qubit bitstrings
    :rtype: list
    """
    return ["".join(labels) for labels in itertools.product('01', repeat=n)]


def sample_bad_readout(program, num_samples, assignment_probs, cxn):
    """
    Generate `n` samples of measuring all outcomes of a Quil `program`
    assuming the assignment probabilities `assignment_probs` by simulating the
    wave function on a qvm QVMConnection `cxn`

    :param pyquil.quil.Program program: The program.
    :param int num_samples: The number of samples
    :param numpy.ndarray assignment_probs: A matrix of assignment probabilities
    :param QVMConnection cxn: the QVM connection.
    :return: The resulting sampled outcomes from assignment_probs applied to cxn, one dimensional.
    :rtype: numpy.ndarray
    """
    wf = cxn.wavefunction(program)
    return sample_outcomes(assignment_probs.dot(abs(wf.amplitudes.ravel())**2), num_samples)


def make_histogram(samples, ksup):
    """
    For a list of samples [s1, s2, ..., sN] taking on integer values from 0 to ksup-1,
    make a histogram of each integer's outcome and return it.

    :param samples: The samples.
    :param ksup: The (exclusive) upper bound
    :return: A histogram of outcomes.
    :rtype: numpy.ndarray
    """
    return np.histogram(samples, np.arange(ksup + 1) - 0.5)[0]


def estimate_assignment_probs(bitstring_prep_histograms):
    """
    Compute the estimated assignment probability matrix for a sequence of single shot histograms
    obtained by running the programs generated by `basis_state_preps()`.

        bitstring_prep_histograms[i,j] = #number of measured outcomes j when running program i

    The assignment probability is obtained by transposing and afterwards normalizing the columns.

        p[j, i] = Probability to measure outcome j when preparing the state with program i.

    :param list|numpy.ndarray bitstring_prep_histograms: A nested list or 2d array with shape
    (d, d), where ``d = 2**nqubits`` is the dimension of the Hilbert space. The first axis varies
    over the state preparation program index, the second axis corresponds to the measured bitstring.
    :return: The assignment probability matrix.
    :rtype: numpy.ndarray
    """
    p = np.array(bitstring_prep_histograms, dtype=float).T
    p /= p.sum(axis=0)[np.newaxis, :]
    return p


def plot_pauli_transfer_matrix(ptransfermatrix, ax, labels, title):
    """
    Visualize the Pauli Transfer Matrix of a process.

    :param numpy.ndarray ptransfermatrix: The Pauli Transfer Matrix
    :param ax: The matplotlib axes.
    :param labels: The labels for the operator basis states.
    :param title: The title for the plot
    :return: The modified axis object.
    :rtype: AxesSubplot

    """
    im = ax.imshow(ptransfermatrix, interpolation="nearest", cmap=rigetti_3_color_cm, vmin=-1,
                   vmax=1)
    dim = len(labels)
    plt.colorbar(im, ax=ax)
    ax.set_xticks(range(dim))
    ax.set_xlabel("Input Pauli Operator", fontsize=20)
    ax.set_yticks(range(dim))
    ax.set_ylabel("Output Pauli Operator", fontsize=20)
    ax.set_title(title, fontsize=25)
    ax.set_xticklabels(labels, rotation=45)
    ax.set_yticklabels(labels)
    ax.grid(False)
    return ax


def generated_states(initial_state, preparations):
    """
    Generate states prepared from channel operators acting on an initial state.
    Typically the channel operators will be unitary.

    :param qutip.Qobj initial_state: The initial state as a density matrix.
    :param (list|tuple) preparations: The unitary channel operators that transform the initial
     state.
    :return: The states generated from preparations acting on intial_state
    :rtype: list
    """
    return [e * initial_state * e.dag() for e in preparations]


def state_histogram(rho, ax=None, title="", threshold=0.001):
    """
    Visualize a density matrix as a 3d bar plot with complex phase encoded
    as the bar color.

    This code is a modified version of
    `an equivalent function in qutip <http://qutip.org/docs/3.1.0/apidoc/functions.html#qutip.visualization.matrix_histogram_complex>`_
    which is released under the (New) BSD license.

    :param qutip.Qobj rho: The density matrix.
    :param Axes3D ax: The axes object.
    :param str title: The axes title.
    :param float threshold: (Optional) minimum magnitude of matrix elements. Values below this
    are hidden.
    :return: The axis
    :rtype: mpl_toolkits.mplot3d.Axes3D
    """
    rho_amps = rho.data.toarray().ravel()
    nqc = int(round(np.log2(rho.shape[0])))
    if ax is None:
        fig = plt.figure(figsize=(10, 6))
        ax = Axes3D(fig, azim=-35, elev=35)
    cmap = rigetti_4_color_cm
    norm = mpl.colors.Normalize(-np.pi, np.pi)
    colors = cmap(norm(np.angle(rho_amps)))
    dzs = abs(rho_amps)
    colors[:, 3] = 1.0 * (dzs > threshold)
    xs, ys = np.meshgrid(range(2 ** nqc), range(2 ** nqc))
    xs = xs.ravel()
    ys = ys.ravel()
    zs = np.zeros_like(xs)
    dxs = dys = np.ones_like(xs) * 0.8

    _ = ax.bar3d(xs, ys, zs, dxs, dys, dzs, color=colors)
    ax.set_xticks(np.arange(2 ** nqc) + .4)
    ax.set_xticklabels(basis_labels(nqc))
    ax.set_yticks(np.arange(2 ** nqc) + .4)
    ax.set_yticklabels(basis_labels(nqc))
    ax.set_zlim3d([0, 1])

    cax, kw = mpl.colorbar.make_axes(ax, shrink=.75, pad=.1)
    cb = mpl.colorbar.ColorbarBase(cax, cmap=cmap, norm=norm)
    cb.set_ticks([-np.pi, -np.pi / 2, 0, np.pi / 2, np.pi])
    cb.set_ticklabels((r'$-\pi$', r'$-\pi/2$', r'$0$', r'$\pi/2$', r'$\pi$'))
    cb.set_label('arg')
    ax.view_init(azim=-55, elev=45)
    ax.set_title(title)
    return ax


def bitlist_to_int(bitlist):
    """Convert a binary bitstring into the corresponding unsigned integer.

    :param list bitlist: A list of ones of zeros.
    :return: The corresponding integer.
    :rtype: int
    """

    ret = 0
    for b in bitlist:
        ret = (ret << 1) | (int(b) & 1)
    return ret


def sample_assignment_probs(qubits, nsamples, cxn):
    """
    Sample the assignment probabilities of qubits using nsamples per measurement, and then compute
    the estimated assignment probability matrix. See the docstring for estimate_assignment_probs for
    more information.

    :param list qubits: Qubits to sample the assignment probabilities for.
    :param int nsamples: The number of samples to use in each measurement.
    :param QPUConnection|QVMConnection cxn: The Connection object to connect to Forest.
    :return: The assignment probability matrix.
    :rtype: numpy.ndarray
    """
    num_qubits = len(qubits)
    dimension = 2 ** num_qubits
    hists = []
    preps = basis_state_preps(*qubits)

    jobs = []
    _log.info('Submitting jobs...')
    for jj, p in izip(TRANGE(dimension), preps):
        jobs.append(cxn.run_and_measure_async(p, qubits, nsamples))

    _log.info('Waiting for results...')
    for jj, job_id in izip(TRANGE(dimension), jobs):
        job = cxn.wait_for_job(job_id)
        results = job.result()
        idxs = list(map(bitlist_to_int, results))
        hists.append(make_histogram(idxs, dimension))
    return estimate_assignment_probs(hists)


def run_in_parallel(programs, nsamples, cxn, shuffle=True):
    """
    Take sequences of Protoquil programs on disjoint qubits and execute a single sequence of
    programs that executes the input programs in parallel. Optionally randomize within each
    qubit-specific sequence.

    The programs are passed as a 2d array of Quil programs, where the (first) outer axis iterates
    over disjoint sets of qubits that the programs involve and the inner axis iterates over a
    sequence of related programs, e.g., tomography sequences, on the same set of qubits.

    :param Union[np.ndarray,List[List[Program]]] programs: A rectangular list of lists, or a 2d
        array of Quil Programs. The outer list iterates over disjoint qubit groups as targets, the
        inner list over programs to run on those qubits, e.g., tomographic sequences.
    :param int nsamples: Number of repetitions for executing each Program.
    :param QPUConnection|QVMConnection cxn: The quantum machine connection.
    :param bool shuffle: If True, the order of each qubit specific sequence (2nd axis) is randomized
        Default is True.
    :return: An array of 2d arrays that provide bitstring histograms for each input program.
        The axis of the outer array iterates over the disjoint qubit groups, the outer axis of the
        inner 2d array iterates over the programs for that group and the inner most axis iterates
        over all possible bitstrings for the qubit group under consideration.
    :rtype np.array
    """

    if shuffle:
        n_groups = len(programs)
        n_progs_per_group = len(programs[0])
        permutations = np.outer(np.ones(n_groups, dtype=int),
                                np.arange(n_progs_per_group, dtype=int))
        inverse_permutations = np.zeros_like(permutations)

        for jj in range(n_groups):
            # in-place operation
            np.random.shuffle(permutations[jj])
            # store inverse permutation
            inverse_permutations[jj] = np.argsort(permutations[jj])

        # apply to programs
        shuffled_programs = np.empty((n_groups, n_progs_per_group), dtype=object)
        for jdx, (progsj, pj) in enumerate(zip(programs, permutations)):
            shuffled_programs[jdx] = [progsj[pjk] for pjk in pj]

        shuffled_results = _run_in_parallel(shuffled_programs, nsamples, cxn)

        # reverse shuffling of results
        results = np.array([resultsj[pj]
                            for resultsj, pj in zip(shuffled_results, inverse_permutations)])
        return results
    else:
        return _run_in_parallel(programs, nsamples, cxn)


def _run_in_parallel(programs, nsamples, cxn):
    """
    See docs for ``run_in_parallel()``.

    :param Union[np.ndarray,List[List[Program]]] programs: A rectangular list of lists, or a 2d
        array of Quil Programs. The outer list iterates over disjoint qubit groups as targets, the
        inner list over programs to run on those qubits, e.g., tomographic sequences.
    :param int nsamples: Number of repetitions for executing each Program.
    :param QPUConnection|QVMConnection cxn: The quantum machine connection.
    :return: An array of 2d arrays that provide bitstring histograms for each input program.
        The axis of the outer array iterates over the disjoint qubit groups, the outer axis of the
        inner 2d array iterates over the programs for that group and the inner most axis iterates
        over all possible bitstrings for the qubit group under consideration. The bitstrings are
        enumerated in lexicographical order, i.e., for a program with qubits {3, 1, 2} the qubits
        are first sorted -> [1, 2, 3] and then the bitstrings are enumerated as 000, 001, 010,
        where the bits ijk correspond to the states of qubits 1,2 and 3, respectively.
    :rtype np.array
    """
    n_groups = len(programs)
    n_progs_per_group = len(programs[0])

    for progs in programs[1:]:
        if not len(progs) == n_progs_per_group:
            raise ValueError("Non-rectangular grid of programs specified: {}".format(programs))

    # identify qubit groups, ensure disjointedness
    qubit_groups = [set() for _ in range(n_groups)]
    for group_idx, group in enumerate(qubit_groups):
        for prog in programs[group_idx]:
            group.update(set(prog.get_qubits()))

        # test that groups are actually disjoint by comparing with the ones already created
        for other_idx, other_group in enumerate(qubit_groups[:group_idx]):
            intersection = other_group & group
            if intersection:
                raise ValueError(
                    "Programs from groups {} and {} intersect on qubits {}".format(
                        other_idx, group_idx, intersection))

    qubit_groups = [sorted(c) for c in qubit_groups]
    all_qubits = sum(qubit_groups, [])
    n_qubits_per_group = [len(c) for c in qubit_groups]

    # create joint programs
    parallel_programs = [sum(progsj, Program()) for progsj in zip(*programs)]

    # execute on cxn
    all_results = []
    for i, prog in izip(TRANGE(n_progs_per_group), parallel_programs):
        try:
            results = cxn.run_and_measure(prog, all_qubits, nsamples)
            all_results.append(np.array(results))
        except QPUError as e:
            _log.error("Could not execute parallel program:\n%s", prog.out())
            raise e

    # generate histograms per qubit group
    all_histograms = np.array([np.zeros((n_progs_per_group, 2 ** n_qubits), dtype=int)
                               for n_qubits in n_qubits_per_group])
    for idx, results in enumerate(all_results):
        n_qubits_seen = 0
        for jdx, n_qubits in enumerate(n_qubits_per_group):
            group_results = results[:, n_qubits_seen:n_qubits_seen + n_qubits]
            outcome_labels = list(map(bitlist_to_int, group_results))
            dimension = 2 ** n_qubits
            all_histograms[jdx][idx] = make_histogram(outcome_labels, dimension)
            n_qubits_seen += n_qubits

    return all_histograms



##############################################################################
# Copyright 2017-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from collections import OrderedDict, namedtuple
import logging
import itertools

from grove.tomography.utils import to_density_matrix, import_qutip

try:
    # Python 2
    from itertools import izip
except ImportError:  # pragma no coverage
    # Python 3
    izip = zip

import numpy as np
from scipy.sparse import hstack as sphstack, vstack as spvstack
from scipy.sparse.linalg import norm as spnorm


_log = logging.getLogger(__name__)

qt = import_qutip()


FROBENIUS = 'fro'
CHOI = 'choi'

EPS = 1e-8


class CRMBaseError(Exception):
    """
    Base class for errors raised when the confusion rate matrix is defective.
    """
    pass


class CRMUnnormalizedError(CRMBaseError):
    """
    Raised when a confusion rate matrix is not properly normalized.
    """
    pass


class CRMValueError(CRMBaseError):
    """
    Raised when a confusion rate matrix contains elements not contained in the interval :math`[0,1]`
    """
    pass


DiagonalPOVM = namedtuple("DiagonalPOVM", ["pi_basis", "confusion_rate_matrix", "ops"])


def make_diagonal_povm(pi_basis, confusion_rate_matrix):
    """
    Create a DiagonalPOVM from a ``pi_basis`` and the ``confusion_rate_matrix`` associated with a
    readout.

    See also the grove documentation.

    :param OperatorBasis pi_basis: An operator basis of rank-1 projection operators.
    :param numpy.ndarray confusion_rate_matrix: The matrix of detection probabilities conditional
    on a prepared qubit state.
    :return: The POVM corresponding to confusion_rate_matrix.
    :rtype: DiagonalPOVM
    """

    confusion_rate_matrix = np.asarray(confusion_rate_matrix)
    if not np.allclose(confusion_rate_matrix.sum(axis=0), np.ones(confusion_rate_matrix.shape[1])):
        raise CRMUnnormalizedError("Unnormalized confusion matrix:\n{}".format(
            confusion_rate_matrix))
    if not (confusion_rate_matrix >= 0).all() or not (confusion_rate_matrix <= 1).all():
        raise CRMValueError("Confusion matrix must have values in [0, 1]:"
                            "\n{}".format(confusion_rate_matrix))

    ops = [sum((pi_j * pjk for (pi_j, pjk) in izip(pi_basis.ops, pjs)), 0)
           for pjs in confusion_rate_matrix]
    return DiagonalPOVM(pi_basis=pi_basis, confusion_rate_matrix=confusion_rate_matrix, ops=ops)


def is_hermitian(operator):
    """
    Check if matrix or operator is hermitian.

    :param (numpy.ndarray|qutip.Qobj) operator: The operator or matrix to be tested.
    :return: True if the operator is hermitian.
    :rtype: bool
    """
    if isinstance(operator, qt.Qobj):
        return (operator.dag() - operator).norm(FROBENIUS) / operator.norm(FROBENIUS) < EPS
    if isinstance(operator, np.ndarray):
        return np.linalg.norm(operator.T.conj() - operator) / np.linalg.norm(operator) < EPS
    return spnorm(operator.H - operator) / spnorm(operator) < EPS


def is_projector(operator):
    """
    Check if operator is a projector.

    :param qutip.Qobj operator: The operator or matrix to be tested.
    :return: True if the operator is a projector.
    :rtype: bool
    """
    # verify that P^dag=P and P^2-P=0 holds up to relative numerical accuracy EPS.
    return (is_hermitian(operator) and (operator * operator - operator).norm(FROBENIUS)
            / operator.norm(FROBENIUS) < EPS)


def choi_matrix(pauli_tm, basis):
    """
    Compute the Choi matrix for a quantum process from its Pauli Transfer Matrix.

    This agrees with the definition in
    `Chow et al. <https://doi.org/10.1103/PhysRevLett.109.060501>`_
    except for a different overall normalization.
    Our normalization agrees with that of qutip.

    :param numpy.ndarray pauli_tm: The Pauli Transfer Matrix as 2d-array.
    :param OperatorBasis basis:  The operator basis, typically products of normalized Paulis.
    :return: The Choi matrix as qutip.Qobj.
    :rtype: qutip.Qobj
    """

    if not basis.is_orthonormal():  # pragma no coverage
        raise ValueError("Need an orthonormal operator basis.")
    if not all((is_hermitian(op) for op in basis.ops)):  # pragma no coverage
        raise ValueError("Need an operator basis of hermitian operators.")

    sbasis = basis.super_basis()
    D = basis.dim
    choi = sum((pauli_tm[jj, kk] * sbasis.ops[jj + kk * D] for jj in range(D) for kk in range(D)))
    choi.superrep = CHOI
    return choi


def to_realimag(z):
    """
    Convert a complex hermitian matrix to a real valued doubled up representation, i.e., for
    ``Z = Z_r + 1j * Z_i`` return ``R(Z)``::

        R(Z) = [ Z_r   Z_i]
               [-Z_i   Z_r]

    A complex hermitian matrix ``Z`` with elementwise real and imaginary parts
    ``Z = Z_r + 1j * Z_i`` can be
    isomorphically represented in doubled up form as::

        R(Z) = [ Z_r   Z_i]
               [-Z_i   Z_r]

        R(X)*R(Y) = [ (X_r*Y_r-X_i*Y_i)    (X_r*Y_i + X_i*Y_r)]
                    [-(X_r*Y_i + X_i*Y_r)  (X_r*Y_r-X_i*Y_i)  ]

                  = R(X*Y).

    In particular, ``Z`` is complex positive (semi-)definite iff ``R(Z)`` is real positive
    (semi-)definite.

    :param (qutip.Qobj|scipy.sparse.base.spmatrix) z:  The operator representation matrix.
    :returns: R(Z) the doubled up representation.
    :rtype: scipy.sparse.csr_matrix
    """
    if isinstance(z, qt.Qobj):
        z = z.data
    if not is_hermitian(z):  # pragma no coverage
        raise ValueError("Need a hermitian matrix z")
    return spvstack([sphstack([z.real, z.imag]), sphstack([z.imag.T, z.real])]).tocsr().real


class OperatorBasis(object):
    """
    Encapsulate a complete set of basis operators.
    """

    def __init__(self, labels_ops):
        """
        Encapsulates a set of linearly independent operators.

        :param (list|tuple) labels_ops: Sequence of tuples (label, operator) where label is a string
            and operator a qutip.Qobj operator representation.
        """
        self.ops_by_label = OrderedDict(labels_ops)
        self.labels = list(self.ops_by_label.keys())
        self.ops = list(self.ops_by_label.values())
        self.dim = len(self.ops)

        # the basis change transformation matrix from a representation in the operator basis
        # to the original basis. We enforce CSR sparse matrix representation to have efficient
        # matrix vector products.
        self.basis_transform = sphstack([qt.operator_to_vector(opj).data
                                         for opj in self.ops]).tocsr()
        self._metric = None
        self._is_orthonormal = None
        self._all_hermitian = None

    def metric(self):
        """
        Compute a matrix of Hilbert-Schmidt inner products for the basis operators, update
        self._metric, and return the value.

        :return: The matrix of inner products.
        :rtype: numpy.matrix
        """
        if self._metric is None:
            _log.debug("Computing and caching operator basis metric")
            self._metric = np.matrix([[(j.dag() * k).tr() for k in self.ops] for j in self.ops])
        return self._metric

    def is_orthonormal(self):
        """
        Compute a matrix of Hilbert-Schmidt inner products for the basis operators, and see if they
        are orthonormal. If they are return True, else, False.

        :return: True if the basis vectors represented by this OperatorBasis are orthonormal, False
         otherwise.
        :rtype: bool
        """
        if self._is_orthonormal is None:
            _log.debug("Testing and caching if operator basis is orthonormal")
            self._is_orthonormal = np.allclose(self.metric(), np.eye(self.dim))
        return self._is_orthonormal

    def all_hermitian(self):
        """
        Check if all basis operators are hermitian.
        """
        if self._all_hermitian is None:
            _log.debug("Testing and caching if all basis operator are hermitian")
            self._all_hermitian = all((is_hermitian(op) for op in self.ops))
        return self._all_hermitian

    def __iter__(self):
        """
        Iterate over tuples of (label, basis_op)

        :return: Yields the labels and qutip operators corresponding to the vectors in this basis.
        :rtype: tuple (str, qutip.qobj.Qobj)
        """
        for l, op in zip(self.labels, self.ops):
            yield l, op

    def product(self, *bases):
        """
        Compute the tensor product with another basis.

        :param bases: One or more additional bases to form the product with.
        :return (OperatorBasis): The tensor product basis as an OperatorBasis object.
        """
        if len(bases) > 1:
            basis_rest = bases[0].product(*bases[1:])
        else:
            assert len(bases) == 1
            basis_rest = bases[0]

        labels_ops = [(b1l + b2l, qt.tensor(b1, b2)) for (b1l, b1), (b2l, b2) in
                      itertools.product(self, basis_rest)]

        return OperatorBasis(labels_ops)

    def __pow__(self, n):
        """
        Create the n-fold tensor product basis.

        :param int n: The number of identical tensor factors.
        :return: The product basis.
        :rtype: OperatorBasis
        """
        if not isinstance(n, int):  # pragma no coverage
            raise TypeError("Can only accept an integer number of factors")
        if n < 1:  # pragma no coverage
            raise ValueError("Need positive number of factors")
        if n == 1:
            return self
        return self.product(*([self] * (n - 1)))

    def super_basis(self):
        """
        Generate the superoperator basis in which the Choi matrix can be represented.

        The follows the definition in
        `Chow et al. <https://doi.org/10.1103/PhysRevLett.109.060501>`_


        :return (OperatorBasis): The super basis as an OperatorBasis object.
        """

        labels_ops = [(bnl + "^T (x) " + bml, qt.sprepost(bm, bn)) for (bnl, bn), (bml, bm) in
                      itertools.product(self, self)]
        return OperatorBasis(labels_ops)

    def project_op(self, op):
        """
        Project an operator onto the basis.

        :param qutip.Qobj op: The operator to project.
        :return: The projection coefficients as a numpy array.
        :rtype: scipy.sparse.csr_matrix
        """
        if not self.is_orthonormal():  # pragma no coverage
            raise ValueError("project_op only implemented for orthonormal operator bases")
        return self.basis_transform.H * qt.operator_to_vector(op).data

    def transfer_matrix(self, superoperator):
        """
        Compute the transfer matrix :math:`R_{jk} = \tr[P_j sop(P_k)]`.

        :param qutip.Qobj superoperator: The superoperator to transform.
        :return: The transfer matrix in sparse form.
        :rtype: scipy.sparse.csr_matrix
        """
        if not self.is_orthonormal():  # pragma no coverage
            raise ValueError("transfer_matrix() only implemented for orthonormal operator bases.")
        return self.basis_transform.H * superoperator.data * self.basis_transform

    def super_from_tm(self, transfer_matrix):
        """
        Reconstruct a super operator from a transfer matrix representation.
        This inverts `self.transfer_matrix(...)`.

        :param (numpy.ndarray) transfer_matrix: A process in transfer matrix form.
        :return: A qutip super operator.
        :rtype: qutip.Qobj.
        """
        if not self.is_orthonormal():  # pragma no coverage
            raise ValueError("super_from_tm() only implemented for orthonormal operator bases")

        data = self.basis_transform * transfer_matrix * self.basis_transform.H
        sop = qt.Qobj(data, dims=[self.ops[0].dims, self.ops[0].dims])
        sop.superrep = "super"
        return sop

    def __repr__(self):
        return "<span[{}]>".format(",".join(self.labels))

    def __eq__(self, other):
        return (self.labels == other.labels and all(
            [(my_op - o_op).norm(FROBENIUS) < EPS for (my_op, o_op) in zip(self.ops, other.ops)]))


if qt:

    # using the Pi-basis for POVM terms allows to easily associate the preparations with individual
    # multi-qubit projectors

    QX = qt.sigmax()
    QY = qt.sigmay()
    QZ = qt.sigmaz()
    QI = qt.qeye(2)
    GS = to_density_matrix(qt.basis(2, 0))
    ES = to_density_matrix(qt.basis(2, 1))
    POVM_PI_BASIS = OperatorBasis([("0", GS), ("1", ES)])
    PAULI_BASIS = OperatorBasis(
        [("I", QI / np.sqrt(2)), ("X", QX / np.sqrt(2)),
         ("Y", QY / np.sqrt(2)), ("Z", QZ / np.sqrt(2))])

else:  # pragma no coverage
    QX = QY = QZ = QI = GS = ES = None
    POVM_PI_BASIS = PAULI_BASIS = None


def n_qubit_pauli_basis(n):
    """
    Construct the tensor product operator basis of `n` PAULI_BASIS's.

    :param int n: The number of qubits.
    :return: The product Pauli operator basis of `n` qubits
    :rtype: OperatorBasis
    """
    if n >= 1:
        return PAULI_BASIS ** n
    else:  # pragma no coverage
        raise ValueError("n = {} should be at least 1.".format(n))


def n_qubit_ground_state(n):
    """
    Construct the tensor product of `n` ground states `|0>`.

    :param int n: The number of qubits.
    :return: The state `|000...0>` for `n` qubits.
    :rtype: qutip.Qobj
    """
    return qt.tensor(*([GS] * n))



from __future__ import division

import numpy as np
from grove.alpha.phaseestimation.phase_estimation import phase_estimation

from grove.alpha.jordan_gradient.gradient_utils import binary_to_real, \
    measurements_to_bf


def gradient_program(f_h, precision):
    """Gradient estimation via Jordan's algorithm
    10.1103/PhysRevLett.95.050501

    :param float f_h: Oracle output at perturbation h.
    :param list ancilla_qubits: Qubits of output register.
    :return: Quil program to estimate gradient of f.
    :rtype: Program
    """

    # encode oracle values into phase
    phase_factor = np.exp(1.0j * 2 * np.pi * abs(f_h))
    U = np.array([[phase_factor, 0],
                  [0, phase_factor]])
    p_gradient = phase_estimation(U, precision)

    return p_gradient


def estimate_gradient(f_h, precision, gradient_max=1, n_measurements=50, 
                      cxn=None):
    """Estimate the gradient using function evaluation at perturbation, h

    :param float f_h: Oracle output at perturbation h.
    :param int precision: Bit precision of gradient.
    :param int gradient_max: OOM estimate of largest gradient value
    :param int n_measurements: Number of times to measure system.
    :param Connection cxn: Connection to the QPU or QVM.
    :return: Decimal estimate of gradient.
    :rtype: float
    """

    # scale f_h by range of values gradient can take on
    f_h *= 1. / gradient_max

    # generate gradient program
    perturbation_sign = np.sign(f_h)
    p_gradient = gradient_program(f_h, precision)

    # run gradient program
    if cxn is None:
        from pyquil.api import QVMConnection
        cxn = QVMConnection()
    measured_qubits = list(range(precision + 1))
    measurements = cxn.run(p_gradient, measured_qubits, n_measurements)

    # summarize measurements
    bf_estimate = perturbation_sign * measurements_to_bf(measurements)
    bf_explicit = '{0:.16f}'.format(bf_estimate)
    deci_estimate = binary_to_real(bf_explicit)
    
    # rescale gradient
    deci_estimate *= gradient_max

    return deci_estimate



"""
Implementation of the swap test

Given two states existing on registers A and B, their overlap can be measured
by performing a swap test.
"""
import numpy as np
from pyquil.quil import Program
from pyquil.gates import H, CSWAP


class RegisterSizeMismatch(Exception):
    """Error associated with mismatched register size"""
    pass


def swap_circuit_generator(register_a, register_b, ancilla):
    """
    Generate the swap test circuit primitive.

    Registers A and B must be of equivalent size for swap to work.  This module
    uses the CSWAP gate in pyquil.

    :param List register_a: qubit labels in the 'A' register
    :param List register_b: qubit labels in the 'B' register
    :param ancilla: ancilla to measure and control the swap operation.
    :return: Program
    """
    if len(register_a) != len(register_b):
        raise RegisterSizeMismatch("registers involve different numbers of qubits")

    if not isinstance(register_a, list):
        raise TypeError("Register A needs to be list")
    if not isinstance(register_b, list):
        raise TypeError("Register B needs to be a list")

    if ancilla is None:
        ancilla = max(register_a + register_b) + 1

    swap_program = Program()
    swap_program += H(ancilla)
    for a, b in zip(register_a, register_b):
        swap_program += CSWAP(ancilla, a, b)
    swap_program += H(ancilla)

    return swap_program


def run_swap_test(program_a, program_b, number_of_measurements, quantum_resource,
                  classical_memory=0, ancilla=None):
    """
    Run a swap test and return the calculated overlap

    :param Program program_a: state preparation for the 'A' register
    :param Program program_b: state preparation for the 'B' register
    :param Int number_of_measurements: number of times to measure
    :param quantum_resource: QAM connection object
    :param Int classical_memory: location of classical memory slot to use
    :param ancilla: Index of ancilla
    :return: overlap of states prepared by program_a and program_b
    :rtype: float
    """
    register_a = list(program_a.get_qubits())
    register_b = list(program_b.get_qubits())
    if ancilla is None:
        ancilla = max(register_a + register_b) + 1

    swap_test_program = Program()
    swap_test_program += program_a + program_b
    swap_test_program += swap_circuit_generator(register_a, register_b, ancilla)
    swap_test_program.measure(ancilla, [classical_memory])

    # TODO: instead of number of measurements have user set precision of overlap
    # estimate and calculate number of shots to be within this precision
    results = quantum_resource.run(swap_test_program,
                                   classical_memory,
                                   trials=number_of_measurements)

    probability_of_one = np.mean(results)
    if probability_of_one > 0.5:
        raise ValueError("measurements indicate overlap is negative")

    estimated_overlap = np.sqrt(1 - 2 * probability_of_one)
    return estimated_overlap








import numpy as np


def binary_to_real(number):
    """Convert binary fraction to real decimal

    :param string number: String representation of binary fraction.
    :return: Real decimal representation of binary fraction.
    :rtype: float
    """

    if isinstance(number, str):
        if number[0] == '-':
            n_sign = -1
        else:
            n_sign = 1
    elif isinstance(number, float):
        n_sign = np.sign(number)
        number = str(number)

    deci = 0
    for ndx, val in enumerate(number.split('.')[-1]):
        deci += float(val) / 2**(ndx+1)
    deci *= n_sign

    return deci


def measurements_to_bf(measurements):
    """Convert measurements into gradient binary fraction

    :param list measurements: Output measurements of gradient program.
    :return: Binary fraction representation of gradient estimate.
    :rtype: float
    """

    measurements = np.array(measurements)
    stats = measurements.sum(axis=0) / len(measurements)
    stats_str = [str(int(i)) for i in np.round(stats[::-1][1:])]
    bf_str = '0.' + ''.join(stats_str)
    bf = float(bf_str)

    return bf



"""
Various ways to group sets of Pauli Terms

This augments the existing infrastructure in pyquil that finds commuting sets
of PauliTerms.
"""
from functools import reduce
from pyquil.paulis import is_identity, PauliTerm, PauliSum


def _commutes(p1, p2):
    # Identity commutes with anything
    if is_identity(p1) or is_identity(p2):
        return True

    # Operators acting on different qubits commute
    if len(set(p1.get_qubits()) & set(p2.get_qubits())) == 0:
        return True

    # Otherwise, they must be the same thing modulo coefficient
    return p1.id() == p2.id()


def diagonal_basis_commutes(pauli_a, pauli_b):
    """
    Test if `pauli_a` and `pauli_b` share a diagonal basis

    Example:

        Check if [A, B] with the constraint that A & B must share a one-qubit
        diagonalizing basis. If the inputs were [sZ(0), sZ(0) * sZ(1)] then this
        function would return True.  If the inputs were [sX(5), sZ(4)] this
        function would return True.  If the inputs were [sX(0), sY(0) * sZ(2)]
        this function would return False.

    :param pauli_a: Pauli term to check commutation against `pauli_b`
    :param pauli_b: Pauli term to check commutation against `pauli_a`
    :return: Boolean of commutation result
    :rtype: Bool
    """
    overlapping_active_qubits = set(pauli_a.get_qubits()) & set(pauli_b.get_qubits())
    for qubit_index in overlapping_active_qubits:
        if (pauli_a[qubit_index] != 'I' and pauli_b[qubit_index] != 'I' and
           pauli_a[qubit_index] != pauli_b[qubit_index]):
            return False

    return True


def get_diagonalizing_basis(list_of_pauli_terms):
    """
    Find the Pauli Term with the most non-identity terms

    :param list_of_pauli_terms: List of Pauli terms to check
    :return: The highest weight Pauli Term
    :rtype: PauliTerm
    """
    qubit_ops = set(reduce(lambda x, y: x + y,
                       [list(term._ops.items()) for term in list_of_pauli_terms]))
    qubit_ops = sorted(list(qubit_ops), key=lambda x: x[0])

    return PauliTerm.from_list(list(map(lambda x: tuple(reversed(x)), qubit_ops)))


def _max_key_overlap(pauli_term, diagonal_sets):
    """
    Calculate the max overlap of a pauli term ID with keys of diagonal_sets

    Returns a different key if we find any collisions.  If no collisions is
    found then the pauli term is added and the key is updated so it has the
    largest weight.

    :param pauli_term:
    :param diagonal_sets:
    :return: dictionary where key value pair is tuple indicating diagonal basis
             and list of PauliTerms that share that basis
    :rtype: dict
    """
    # a lot of the ugliness comes from the fact that
    # list(PauliTerm._ops.items()) is not the appropriate input for
    # Pauliterm.from_list()
    for key in list(diagonal_sets.keys()):
        pauli_from_key = PauliTerm.from_list(
            list(map(lambda x: tuple(reversed(x)), key)))
        if diagonal_basis_commutes(pauli_term, pauli_from_key):
            updated_pauli_set = diagonal_sets[key] + [pauli_term]
            diagonalizing_term = get_diagonalizing_basis(updated_pauli_set)
            if len(diagonalizing_term) > len(key):
                del diagonal_sets[key]
                new_key = tuple(sorted(diagonalizing_term._ops.items(),
                                       key=lambda x: x[0]))
                diagonal_sets[new_key] = updated_pauli_set
            else:
                diagonal_sets[key] = updated_pauli_set
            return diagonal_sets
    # made it through all keys and sets so need to make a new set
    else:
        # always need to sort because new pauli term functionality
        new_key = tuple(sorted(pauli_term._ops.items(), key=lambda x: x[0]))
        diagonal_sets[new_key] = [pauli_term]
        return diagonal_sets


def commuting_sets_by_zbasis(pauli_sums):
    """
    Computes commuting sets based on terms having the same diagonal basis

    Following the technique outlined in the appendix of arXiv:1704.05018.

    :param pauli_sums: PauliSum object to group
    :return: dictionary where key value pair is a tuple corresponding to the
             basis and a list of PauliTerms associated with that basis.
    """
    diagonal_sets = {}
    for term in pauli_sums:
        diagonal_sets = _max_key_overlap(term, diagonal_sets)

    return diagonal_sets


def check_trivial_commutation(pauli_list, single_pauli_term):
    """
    Check if a PauliTerm trivially commutes with a list of other terms.

    :param list pauli_list: A list of PauliTerm objects
    :param PauliTerm single_pauli_term: A PauliTerm object
    :returns: True if pauli_two object commutes with pauli_list, False otherwise
    :rtype: bool
    """
    if not isinstance(pauli_list, list):
        raise TypeError("pauli_list should be a list")

    for term in pauli_list:
        if not _commutes(term, single_pauli_term):
            return False
    return True


def commuting_sets_by_indices(pauli_sums, commutation_check):
    """
    For a list of pauli sums, find commuting sets and keep track of which pauli sum they came from.

    :param pauli_sums: A list of PauliSum
    :param commutation_check: a function that checks if all elements of a list
                              and a single pauli term commute.
    :return: A list of commuting sets. Each set is a list of tuples (i, j) to find the particular
        commuting term. i is the index of the pauli sum from whence the term came. j is the
        index within the set.
    """
    assert isinstance(pauli_sums, list)

    group_inds = []
    group_terms = []
    for i, pauli_sum in enumerate(pauli_sums):
        for j, term in enumerate(pauli_sum):
            if len(group_inds) == 0:
                # Initialization
                group_inds.append([(i, j)])
                group_terms.append([term])
                continue

            for k, group in enumerate(group_terms):
                if commutation_check(group, term):
                    group_inds[k] += [(i, j)]
                    group_terms[k] += [term]
                    break
            else:
                # for ... else means loop completed without a `break`
                # Which means this needs to start its own group.
                group_inds.append([(i, j)])
                group_terms.append([term])

    return group_inds


def commuting_sets_trivial(pauli_sum):
    """
    Group a pauli term into commuting sets using trivial check

    :param pauli_sum: PauliSum term
    :return: list of lists containing individual Pauli Terms
    """
    if not isinstance(pauli_sum, (PauliTerm, PauliSum)):
        raise TypeError("This method can only group PauliTerm or PauliSum objects")

    if isinstance(pauli_sum, PauliTerm):
        pauli_sum = PauliSum([pauli_sum])

    commuting_terms = []
    for term in pauli_sum:
        # find the group that it trivially commutes with
        for term_group in commuting_terms:
            if check_trivial_commutation(term_group, term):
                term_group.append(term)
                break
        else:
            commuting_terms.append([term])

    return commuting_terms




"""
Utilities for estimating expected values of Pauli terms given pyquil programs
"""
from collections import namedtuple

import numpy as np
from pyquil.paulis import (PauliSum, PauliTerm, commuting_sets, sI,
                           term_with_coeff, is_identity, is_zero)
from pyquil.quil import Program
from pyquil.gates import RY, RX, MEASURE
from grove.measurements.term_grouping import commuting_sets_by_zbasis


class CommutationError(Exception):
    pass


def remove_imaginary_terms(pauli_sums):
    """
    Remove the imaginary component of each term in a Pauli sum

    :param PauliSum pauli_sums: The Pauli sum to process.
    :return: a purely hermitian Pauli sum.
    :rtype: PauliSum
    """
    if not isinstance(pauli_sums, PauliSum):
        raise TypeError("not a pauli sum. please give me one")
    new_term = sI(0) * 0.0
    for term in pauli_sums:
        new_term += term_with_coeff(term, term.coefficient.real)

    return new_term


def get_rotation_program(pauli_term):
    """
    Generate a rotation program so that the pauli term is diagonal

    :param PauliTerm pauli_term: The Pauli term used to generate diagonalizing
                                 one-qubit rotations.
    :return: The rotation program.
    :rtype: Program
    """
    meas_basis_change = Program()
    for index, gate in pauli_term:
        if gate == 'X':
            meas_basis_change.inst(RY(-np.pi / 2, index))
        elif gate == 'Y':
            meas_basis_change.inst(RX(np.pi / 2, index))
        elif gate == 'Z':
            pass
        else:
            raise ValueError()

    return meas_basis_change


def get_parity(pauli_terms, bitstring_results):
    """
    Calculate the eigenvalues of Pauli operators given results of projective measurements

    The single-qubit projective measurement results (elements of
    `bitstring_results`) are listed in physical-qubit-label numerical order.

    An example:

    Consider a Pauli term Z1 Z5 Z6 I7 and a collection of single-qubit
    measurement results corresponding to measurements in the z-basis on qubits
    {1, 5, 6, 7}. Each element of bitstring_results is an element of
    :math:`\{0, 1\}^{\otimes 4}`.  If [0, 0, 1, 0] and [1, 0, 1, 1]
    are the two projective measurement results in `bitstring_results` then
    this method returns a 1 x 2 numpy array with values [[-1, 1]]

    :param List pauli_terms: A list of Pauli terms operators to use
    :param bitstring_results: A list of projective measurement results.  Each
                              element is a list of single-qubit measurements.
    :return: Array (m x n) of {+1, -1} eigenvalues for the m-operators in
             `pauli_terms` associated with the n measurement results.
    :rtype: np.ndarray
    """
    qubit_set = []
    for term in pauli_terms:
        qubit_set.extend(list(term.get_qubits()))
    active_qubit_indices = sorted(list(set(qubit_set)))
    index_mapper = dict(zip(active_qubit_indices,
                            range(len(active_qubit_indices))))

    results = np.zeros((len(pauli_terms), len(bitstring_results)))

    # convert to array so we can fancy index into it later.
    # list() is needed to cast because we don't really want a map object
    bitstring_results = list(map(np.array, bitstring_results))
    for row_idx, term in enumerate(pauli_terms):
        memory_index = np.array(list(map(lambda x: index_mapper[x],
                                         sorted(term.get_qubits()))))

        results[row_idx, :] = [-2 * (sum(x[memory_index]) % 2) +
                               1 for x in bitstring_results]
    return results


EstimationResult = namedtuple('EstimationResult',
                              ('expected_value', 'pauli_expectations',
                               'covariance', 'variance', 'n_shots'))


def estimate_pauli_sum(pauli_terms, basis_transform_dict, program,
                       variance_bound, quantum_resource,
                       commutation_check=True):
    """
    Estimate the mean of a sum of pauli terms to set variance

    The sample variance is calculated by

    .. math::
        \begin{align}
        \mathrm{Var}[\hat{\langle H \rangle}] = \sum_{i, j}h_{i}h_{j}
        \mathrm{Cov}(\hat{\langle P_{i} \rangle}, \hat{\langle P_{j} \rangle})
        \end{align}

    The expectation value of each Pauli operator (term and coefficient) is
    also returned.  It can be accessed through the named-tuple field
    `pauli_expectations'.

    :param pauli_terms: list of pauli terms to measure simultaneously or a
                        PauliSum object
    :param basis_transform_dict: basis transform dictionary where the key is
                                 the qubit index and the value is the basis to
                                 rotate into. Valid basis is [I, X, Y, Z].
    :param program: program generating a state to sample from
    :param variance_bound:  Bound on the variance of the estimator for the
                            PauliSum. Remember this is the SQUARE of the
                            standard error!
    :param quantum_resource: quantum abstract machine object
    :param Bool commutation_check: Optional flag toggling a safety check
                                   ensuring all terms in `pauli_terms`
                                   commute with each other
    :return: estimated expected value, expected value of each Pauli term in
             the sum, covariance matrix, variance of the estimator, and the
             number of shots taken.  The objected returned is a named tuple with
             field names as follows: expected_value, pauli_expectations,
             covariance, variance, n_shots.
             `expected_value' == coef_vec.dot(pauli_expectations)
    :rtype: EstimationResult
    """
    if not isinstance(pauli_terms, (list, PauliSum)):
        raise TypeError("pauli_terms needs to be a list or a PauliSum")

    if isinstance(pauli_terms, PauliSum):
        pauli_terms = pauli_terms.terms

    # check if each term commutes with everything
    if commutation_check:
        if len(commuting_sets(sum(pauli_terms))) != 1:
            raise CommutationError("Not all terms commute in the expected way")

    pauli_for_rotations = PauliTerm.from_list(
        [(value, key) for key, value in basis_transform_dict.items()])

    program += get_rotation_program(pauli_for_rotations)

    qubits = sorted(list(basis_transform_dict.keys()))
    for qubit in qubits:
        program.inst(MEASURE(qubit, qubit))

    coeff_vec = np.array(
        list(map(lambda x: x.coefficient, pauli_terms))).reshape((-1, 1))

    # upper bound on samples given by IV of arXiv:1801.03524
    num_sample_ubound = int(np.ceil(np.sum(np.abs(coeff_vec))**2 / variance_bound))
    results = None
    sample_variance = np.infty
    number_of_samples = 0
    while (sample_variance > variance_bound and
           number_of_samples < num_sample_ubound):
        tresults = quantum_resource.run(program, qubits, trials=min(10000, num_sample_ubound))
        number_of_samples += len(tresults)

        parity_results = get_parity(pauli_terms, tresults)

        # Note: easy improvement would be to update mean and variance on the fly
        # instead of storing all these results.
        if results is None:
            results = parity_results
        else:
            results = np.hstack((results, parity_results))

        # calculate the expected values....
        covariance_mat = np.cov(results)
        sample_variance = coeff_vec.T.dot(covariance_mat).dot(coeff_vec) / results.shape[1]

    return EstimationResult(expected_value=coeff_vec.T.dot(np.mean(results, axis=1)),
                            pauli_expectations=np.multiply(coeff_vec.flatten(), np.mean(results, axis=1).flatten()),
                            covariance=covariance_mat,
                            variance=sample_variance,
                            n_shots=results.shape[1])


def remove_identity(psum):
    """
    Remove the identity term from a Pauli sum

    :param psum:
    :return:
    """
    new_psum = []
    identity_terms = []
    for term in psum:
        if not is_identity(term):
            new_psum.append(term)
        else:
            identity_terms.append(term)
    return sum(new_psum), sum(identity_terms)


def estimate_locally_commuting_operator(program, pauli_sum,
                                                 variance_bound,
                                                 quantum_resource):
    """
    Estimate the expected value of a Pauli sum to fixed precision

    :param program: state preparation program
    :param pauli_sum: pauli sum of operators to estimate expected value
    :param variance_bound: variance bound on the estimator
    :param quantum_resource: quantum abstract machine object
    :return: expected value, estimator variance, total number of experiments
    """
    pauli_sum, identity_term = remove_identity(pauli_sum)
    psets = commuting_sets_by_zbasis(pauli_sum)
    variance_bound_per_set = variance_bound / len(psets)

    if isinstance(identity_term, int):
        if np.isclose(identity_term, 0):
            expected_value = 0
        else:
            expected_value = identity_term

    elif isinstance(identity_term, (PauliTerm, PauliSum)):
        if isinstance(identity_term, PauliTerm):
            expected_value = identity_term.coefficient
        else:
            expected_value = identity_term[0].coefficient
    else:
        print(type(identity_term))
        raise TypeError("identity_term must be a PauliTerm or integer")

    total_shots = 0
    estimator_variance = 0

    for qubit_op_key, pset in psets.items():
        results = estimate_pauli_sum(pset, dict(qubit_op_key), program,
                                     variance_bound_per_set,
                                     quantum_resource,
                                     commutation_check=False)

        expected_value += results.expected_value
        total_shots += results.n_shots
        estimator_variance += results.variance

    return expected_value, estimator_variance, total_shots



##############################################################################
# Copyright 2017-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

"""
Module for quantum state and process tomography.

Quantum state and process tomography are algorithms that take as input many copies of a quantum
state or process, and output an estimate of what that state or process is. For more information, see
the documentation.
"""

from __future__ import print_function

import logging
from collections import namedtuple, OrderedDict
from itertools import product as cartesian_product

from pyquil.quilbase import Pragma

import grove.tomography.operator_utils

try:
    # Python 2
    from itertools import izip
except ImportError:  # pragma no coverage
    # Python 3
    izip = zip

import numpy as np

from pyquil.gates import I, RX, RY, MEASURE
from pyquil.quil import Program

import grove.tomography.utils as ut
import grove.tomography.operator_utils as o_ut
from grove.tomography.operator_utils import QI, QX, QY
from grove.tomography.utils import bitlist_to_int, sample_assignment_probs

_log = logging.getLogger(__name__)

qt = ut.import_qutip()
cvxpy = ut.import_cvxpy()

# We constrain the number of qubits to prevent the use of large amounts of memory, and prohibitively
# long programs.
MAX_QUBITS_STATE_TOMO = 4
MAX_QUBITS_PROCESS_TOMO = MAX_QUBITS_STATE_TOMO // 2
SEED = 137
SOLVER = "SCS"
if qt:
    TOMOGRAPHY_GATES = OrderedDict([(I, QI),
                                    (lambda q: RX(np.pi / 2, q), (-1j * np.pi / 4 * QX).expm()),
                                    (lambda q: RY(np.pi / 2, q), (-1j * np.pi / 4 * QY).expm()),
                                    (lambda q: RX(np.pi, q), (-1j * np.pi / 2 * QX).expm())])
else:  # pragma no coverage
    TOMOGRAPHY_GATES = {}


def default_rotations(*qubits):
    """
    Generates the Quil programs for the tomographic pre- and post-rotations of any number of qubits.

    :param list qubits: A list of qubits to perform tomography on.
    """
    for gates in cartesian_product(TOMOGRAPHY_GATES.keys(), repeat=len(qubits)):
        tomography_program = Program()
        for qubit, gate in izip(qubits, gates):
            tomography_program.inst(gate(qubit))
        yield tomography_program


def default_channel_ops(nqubits):
    """
    Generate the tomographic pre- and post-rotations of any number of qubits as qutip operators.

    :param int nqubits: The number of qubits to perform tomography on.
    :return: Qutip object corresponding to the tomographic rotation.
    :rtype: Qobj
    """
    for gates in cartesian_product(TOMOGRAPHY_GATES.values(), repeat=nqubits):
        yield qt.tensor(*gates)


class _SDP_SOLVER(object):
    """
    Helper object that allows to test whether a working convex solver with SDP capabilities is
    installed. Not all solvers supported by cvxpy support positivity constraints. Examples of ones
    that do are CVXOPT and SCS.

    Usage:
        if _SDP_SOLVER.is_functional():
            # code to solve SDP
    """
    _functional = False
    _tested = False

    @classmethod
    def is_functional(cls):
        """
        Checks lazily whether a convex solver is installed that handles positivity constraints.

        :return: True if a solver supporting positivity constraints is installed.
        :rtype: bool
        """
        if not cls._tested:
            cls._tested = True
            np.random.seed(SEED)
            test_problem_dimension = 10
            mat = np.random.randn(test_problem_dimension, test_problem_dimension)
            posmat = mat.dot(mat.T)
            posvar = cvxpy.Variable(test_problem_dimension, test_problem_dimension)
            prob = cvxpy.Problem(cvxpy.Minimize((cvxpy.trace(posmat * posvar)
                                                 + cvxpy.norm(posvar))),
                                 [posvar >> 0, cvxpy.trace(posvar) >= 1.])

            try:
                prob.solve(SOLVER)
                cls._functional = True
            except cvxpy.SolverError:  # pragma no coverage
                _log.warning("No convex SDP solver found. You will not be able to solve"
                             " tomography problems with matrix positivity constraints.")
        return cls._functional


TomographySettings = namedtuple('TomographySettings', ('constraints', 'solver_kwargs'))
"""
Encapsulate the TomographySettings, i.e., the constraints to be applied to the Maximum
Likelihood Estimation and the keyword arguments to be passed to the convex solver.

:param set constraints: The constraints to be applied:
For state tomography the maximal constraints are `{'positive', 'unit_trace'}`.
For process tomography the maximal constraints are `{'cpositive', 'trace_preserving'}`.
:param dict solver_kwargs: Keyword arguments to be passed to the convex solver.
"""


DEFAULT_SOLVER_KWARGS = dict(verbose=False, max_iters=20000)


class TomographyBaseError(Exception):
    """
    Base class for errors raised during Tomography analysis.
    """
    pass


class IncompleteTomographyError(TomographyBaseError):
    """
    Raised when a tomography SignalTensor has circuit results that are all 0. indicating that the
    measurement did not complete successfully.
    """
    pass


class BadReadoutPOVM(TomographyBaseError):
    """
    Raised when the tomography analysis fails due to a bad readout calibration.
    """
    pass


class TomographyBase(object):
    def estimate_from_ssr(self):  # pragma no coverage
        raise NotImplementedError()


def _do_tomography(target_program, nsamples, cxn, qubits, max_num_qubits, tomography_class,
                   program_generator, settings, use_run=False):
    """

    :param Program target_program: The program to run to generate the state or process.
    :param int nsamples: The number of samples to take per measurement.
    :param QPUConnection|QVMConnection cxn: The connection to Forest.
    :param lists qubits: The list of qubits to perform tomography on.
    :param int max_num_qubits: The maximum allowed number of qubits.
    :param type tomography_class: The type of tomography to perform.
    :param function program_generator: The function that yields the tomography experiments.
    :param TomographySettings settings: The settings for running the optimizer.
    :param bool use_run: If ``True``, use append measurements on all qubits and use ``cxn.run``
        instead of ``cxn.run_and_measure``.
    :return: The tomography result, the assignment probabilities, and the histograms of counts
     measured.
    :rtype: tuple
    """
    if qubits is None:
        qubits = sorted(target_program.get_qubits())

    num_qubits = len(qubits)
    dimension = 2 ** num_qubits

    if num_qubits > max_num_qubits:
        raise ValueError("Too many qubits!")

    assignment_probs = sample_assignment_probs(qubits, nsamples, cxn)
    tomo_seq = list(program_generator(target_program, qubits))
    histograms = np.zeros((len(tomo_seq), dimension))

    jobs = []
    _log.info('Submitting jobs...')
    for i, tomo_prog in izip(ut.TRANGE(len(tomo_seq)), tomo_seq):
        if use_run:
            jobs.append(cxn.run_async(tomo_prog + Program([MEASURE(q, q) for q in qubits]),
                                      qubits, nsamples))
        else:
            jobs.append(cxn.run_and_measure_async(tomo_prog, qubits, nsamples))
    _log.info('Waiting for results...')
    for i, job_id in izip(ut.TRANGE(len(jobs)), jobs):
        job = cxn.wait_for_job(job_id)
        results = job.result()
        idxs = list(map(bitlist_to_int, results))
        histograms[i] = ut.make_histogram(idxs, dimension)

    povm = o_ut.make_diagonal_povm(grove.tomography.operator_utils.POVM_PI_BASIS ** num_qubits, assignment_probs)
    channel_ops = list(default_channel_ops(num_qubits))

    # Currently the analysis pathways are slightly different, so we branch on which type of
    # tomography is being done.
    if tomography_class.__tomography_type__ == "PROCESS":
        histograms = histograms.reshape((len(channel_ops), len(channel_ops), dimension))
        tomo_result = tomography_class.estimate_from_ssr(histograms, povm, channel_ops, channel_ops,
                                                         settings)
    else:
        tomo_result = tomography_class.estimate_from_ssr(histograms, povm, channel_ops, settings)
    return tomo_result, assignment_probs, histograms



##############################################################################
# Copyright 2017-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import logging

import numpy as np
import matplotlib.pyplot as plt
from pyquil.quilbase import Pragma
from scipy.sparse import (vstack as spvstack, csr_matrix, kron as spkron)
from pyquil.quil import Program

import grove.tomography.operator_utils
from grove.tomography.tomography import TomographyBase, TomographySettings, DEFAULT_SOLVER_KWARGS
import grove.tomography.state_tomography as state_tomography
from grove.tomography import tomography
import grove.tomography.utils as ut
import grove.tomography.operator_utils as o_ut

_log = logging.getLogger(__name__)


qt = ut.import_qutip()
cvxpy = ut.import_cvxpy()


TRACE_PRESERVING = 'trace_preserving'
COMPLETELY_POSITIVE = 'cpositive'
DEFAULT_PROCESS_TOMO_SETTINGS = TomographySettings(
    constraints=set([TRACE_PRESERVING]),
    solver_kwargs=DEFAULT_SOLVER_KWARGS
)


def _prepare_b_jkl_mn(readout_povm, pauli_basis, pre_channel_ops, post_channel_ops, rho0):
    """
    Prepare the coefficient matrix for process tomography. This function uses sparse matrices
    for much greater efficiency. The coefficient matrix is defined as:

    .. math::

            B_{(jkl)(mn)}=\sum_{r,q}\pi_{jr}(\mathcal{R}_{k})_{rm} (\mathcal{R}_{l})_{nq} (\rho_0)_q

    where :math:`\mathcal{R}_{k}` is the transfer matrix of the quantum map corresponding to the
    k-th pre-measurement channel, while :math:`\mathcal{R}_{l}` is the transfer matrix of the l-th
    state preparation process. We also require the overlap
    between the (generalized) Pauli basis ops and the projection operators
    :math:`\pi_{jl}:=\sbraket{\Pi_j}{P_l} = \tr{\Pi_j P_l}`.

    See the grove documentation on tomography for detailed information.

    :param DiagonalPOVM readout_povm: The POVM corresponding to the readout plus classifier.
    :param OperatorBasis pauli_basis: The (generalized) Pauli basis employed in the estimation.
    :param list pre_channel_ops: The state preparation channel operators as `qutip.Qobj`
    :param list post_channel_ops: The pre-measurement (post circuit) channel operators as `qutip.Qobj`
    :param qutip.Qobj rho0: The initial state as a density matrix.
    :return: The coefficient matrix necessary to set up the binomial state tomography problem.
    :rtype: scipy.sparse.csr_matrix
    """
    c_jk_m = state_tomography._prepare_c_jk_m(readout_povm, pauli_basis, post_channel_ops)
    pre_channel_transfer_matrices = [pauli_basis.transfer_matrix(qt.to_super(ek))
                                     for ek in pre_channel_ops]
    rho0_q = pauli_basis.project_op(rho0)

    # These next lines hide some very serious (sparse-)matrix index magic,
    # basically we exploit the same index math as in `qutip.sprepost()`
    # i.e., if a matrix X is linearly mapped `X -> A.dot(X).dot(B)`
    # then this can be rewritten as
    #           `np.kron(B.T, A).dot(X.T.ravel()).reshape((B.shape[1], A.shape[0])).T`
    # The extra matrix transpose operations are necessary because numpy by default
    # uses row-major storage, whereas these operations are conventionally defined for column-major
    # storage.
    d_ln = spvstack([(rlnq * rho0_q).T for rlnq in pre_channel_transfer_matrices]).tocoo()
    b_jkl_mn = spkron(d_ln, c_jk_m).real
    return b_jkl_mn


class ProcessTomography(TomographyBase):
    """
    A ProcessTomography object encapsulates the result of quantum process estimation from
    tomographic data. It provides convenience functions for visualization, computing process
    fidelities and inter-conversion between different numerical representation of quantum processes.
    """
    __tomography_type__ = "PROCESS"

    @staticmethod
    def estimate_from_ssr(histograms, readout_povm, pre_channel_ops, post_channel_ops, settings):
        """
        Estimate a quantum process from single shot histograms obtained by preparing specific input
        states and measuring bitstrings in the Z-eigenbasis after application of given channel
        operators.

        :param numpy.ndarray histograms: The single shot histograms.
        :param DiagonalPOVM readout_povm: The POVM corresponding to readout plus classifier.
        :param list pre_channel_ops: The input state preparation channels as `qutip.Qobj`'s.
        :param list post_channel_ops: The tomography post-process channels as `qutip.Qobj`'s.
        :param TomographySettings settings: The solver and estimation settings.
        :return: The ProcessTomography object and results from the the given data.
        :rtype: ProcessTomography
        """
        nqc = len(pre_channel_ops[0].dims[0])
        pauli_basis = grove.tomography.operator_utils.PAULI_BASIS ** nqc
        pi_basis = readout_povm.pi_basis

        if not histograms.shape[-1] == pi_basis.dim:  # pragma no coverage
            raise ValueError("Currently tomography is only implemented for two-level systems")

        rho0 = grove.tomography.operator_utils.n_qubit_ground_state(nqc)

        n_lkj = np.asarray(histograms)

        b_jkl_mn = _prepare_b_jkl_mn(readout_povm, pauli_basis, pre_channel_ops,
                                     post_channel_ops, rho0)

        r_mn = cvxpy.Variable(pauli_basis.dim ** 2)
        p_jkl = b_jkl_mn.real * r_mn
        obj = -np.matrix(n_lkj.ravel()) * cvxpy.log(p_jkl)

        # cvxpy has col-major order and we collapse k and l onto single dimension
        p_jkl_mat = cvxpy.reshape(p_jkl, pi_basis.dim, len(pre_channel_ops) * len(post_channel_ops))

        # Default constraints:
        # MLE must describe valid probability distribution
        # i.e., for each k and l, p_jkl must sum to one and be element-wise non-negative:
        # 1. \sum_j p_jkl == 1  for all k, l
        # 2. p_jkl >= 0         for all j, k, l
        # where p_jkl = \sum_m b_jkl_mn r_mn
        constraints = [p_jkl >= 0,
                       np.matrix(np.ones((1, pi_basis.dim))) * p_jkl_mat == 1]
        r_mn_mat = cvxpy.reshape(r_mn, pauli_basis.dim, pauli_basis.dim)
        super_pauli_basis = pauli_basis.super_basis()
        choi_real_imag = sum((r_mn_mat[jj, kk] * o_ut.to_realimag(
            super_pauli_basis.ops[jj + kk * pauli_basis.dim])
                              for jj in range(pauli_basis.dim)
                              for kk in range(pauli_basis.dim)), 0)

        if COMPLETELY_POSITIVE in settings.constraints:
            if tomography._SDP_SOLVER.is_functional():
                constraints.append(choi_real_imag >> 0)
            else:  # pragma no coverage
                _log.warning("No convex solver capable of semi-definite problems installed.\n"
                             "Dropping the complete positivity constraint on the process")

        if TRACE_PRESERVING in settings.constraints:
            constraints.append(r_mn_mat[0, 0] == 1)
            constraints.append(r_mn_mat[0, 1:] == 0)

        prob = cvxpy.Problem(cvxpy.Minimize(obj), constraints)
        _ = prob.solve(solver=tomography.SOLVER, **settings.solver_kwargs)

        r_mn_est = r_mn.value.reshape((pauli_basis.dim, pauli_basis.dim)).transpose()
        return ProcessTomography(r_mn_est, pauli_basis, settings)

    def __init__(self, r_est, pauli_basis, settings):
        """
        Construct a ProcessTomography to encapsulate the result of estimating a quantum process
        from a quantum tomography measurement.

        :param numpy.ndarray r_est: The estimated quantum process represented as a Pauli transfer
         matrix.
        :param OperatorBasis pauli_basis: The employed (generalized) Pauli basis.
        :param TomographySettings settings: The settings used to estimate the process.
        """
        self.r_est = r_est
        self.sop = pauli_basis.super_from_tm(r_est)
        self.pauli_basis = pauli_basis
        self.settings = settings

    def process_fidelity(self, reference_unitary):
        """
        Compute the quantum process fidelity of the estimated state with respect to a unitary
        process. For non-sparse reference_unitary, this implementation this will be expensive in
        higher dimensions.

        :param (qutip.Qobj|matrix-like) reference_unitary: A unitary operator that induces a process
         as ``rho -> other*rho*other.dag()``, can also be a superoperator or Pauli-transfer matrix.
        :return: The process fidelity, a real number between 0 and 1.
        :rtype: float
        """
        if isinstance(reference_unitary, qt.Qobj):
            if not reference_unitary.issuper or reference_unitary.superrep != "super":
                sother = qt.to_super(reference_unitary)
            else:
                sother = reference_unitary
            tm_other = self.pauli_basis.transfer_matrix(sother)
        else:
            tm_other = csr_matrix(reference_unitary)
        dimension = self.pauli_basis.ops[0].shape[0]
        return np.trace(tm_other.T * self.r_est).real / dimension ** 2

    def avg_gate_fidelity(self, reference_unitary):
        """
        Compute the average gate fidelity of the estimated process with respect to a unitary
         process. See `Chow et al., 2012, <https://doi.org/10.1103/PhysRevLett.109.060501>`_

        :param (qutip.Qobj|matrix-like) reference_unitary: A unitary operator that induces a process
         as `rho -> other*rho*other.dag()`, alternatively a superoperator or Pauli-transfer matrix.
        :return: The average gate fidelity, a real number between 1/(d+1) and 1, where d is the
        Hilbert space dimension.
        :rtype: float
        """
        process_fidelity = self.process_fidelity(reference_unitary)
        dimension = self.pauli_basis.ops[0].shape[0]
        return (dimension * process_fidelity + 1.0) / (dimension + 1.0)

    def to_super(self):
        """
        Compute the standard superoperator representation of the estimated process.

        :return: The process as a superoperator.
        :rytpe: qutip.Qobj
        """
        return self.sop

    def to_choi(self):
        """
        Compute the choi matrix representation of the estimated process.

        :return: The process as a choi-matrix.
        :rytpe: qutip.Qobj
        """
        return qt.to_choi(self.sop)

    def to_chi(self):
        """
        Compute the chi process matrix representation of the estimated process.

        :return: The process as a chi-matrix.
        :rytpe: qutip.Qobj
        """
        return qt.to_chi(self.sop)

    def to_kraus(self):
        """
        Compute the Kraus operator representation of the estimated process.

        :return: The process as a list of Kraus operators.
        :rytpe: List[np.array]
        """
        return [k.data.toarray() for k in qt.to_kraus(self.sop)]

    def plot_pauli_transfer_matrix(self, ax):
        """
        Plot the elements of the Pauli transfer matrix.

        :param matplotlib.Axes ax: A matplotlib Axes object to plot into.
        """
        title = "Estimated process"
        ut.plot_pauli_transfer_matrix(self.r_est, ax, self.pauli_basis.labels, title)

    def plot(self):
        """
        Visualize the process.

        :return: The generated figure.
        :rtype: matplotlib.Figure
        """
        fig, (ax1) = plt.subplots(1, 1, figsize=(10, 8))
        self.plot_pauli_transfer_matrix(ax1)
        return fig


def process_tomography_programs(process, qubits=None,
                                pre_rotation_generator=tomography.default_rotations,
                                post_rotation_generator=tomography.default_rotations):
    """
    Generator that yields tomographic sequences that wrap a process encoded by a QUIL program `proc`
    in tomographic rotations on the specified `qubits`.

    If `qubits is None`, it assumes all qubits in the program should be
    tomographically rotated.

    :param Program process: A Quil program
    :param list|NoneType qubits: The specific qubits for which to generate the tomographic sequences
    :param pre_rotation_generator: A generator that yields tomographic pre-rotations to perform.
    :param post_rotation_generator: A generator that yields tomographic post-rotations to perform.
    :return: Program for process tomography.
    :rtype: Program
    """
    if qubits is None:
        qubits = process.get_qubits()
    for tomographic_pre_rotation in pre_rotation_generator(*qubits):
        for tomography_post_rotation in post_rotation_generator(*qubits):
            process_tomography_program = Program(Pragma("PRESERVE_BLOCK"))
            process_tomography_program.inst(tomographic_pre_rotation)
            process_tomography_program.inst(process)
            process_tomography_program.inst(tomography_post_rotation)
            process_tomography_program.inst(Pragma("END_PRESERVE_BLOCK"))

            yield process_tomography_program


def do_process_tomography(process, nsamples, cxn, qubits=None, use_run=False):
    """
    Method to perform a process tomography.

    :param Program process: Process to execute.
    :param int nsamples: Number of samples to take for the program.
    :param QVMConnection|QPUConnection cxn: Connection on which to run the program.
    :param list qubits: List of qubits for the program.
    to use in the tomography analysis.
    :param bool use_run: If ``True``, use append measurements on all qubits and use ``cxn.run``
        instead of ``cxn.run_and_measure``.
    :return: The process tomogram
    :rtype: ProcessTomography
    """
    return tomography._do_tomography(process, nsamples, cxn, qubits,
                                     tomography.MAX_QUBITS_PROCESS_TOMO,
                                     ProcessTomography, process_tomography_programs,
                                     DEFAULT_PROCESS_TOMO_SETTINGS, use_run=use_run)



##############################################################################
# Copyright 2017-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import logging

import numpy as np
import matplotlib.pyplot as plt
from pyquil.quilbase import Pragma
from scipy.sparse import csr_matrix, coo_matrix
from pyquil.quil import Program

import grove.tomography.operator_utils
from grove.tomography.tomography import TomographyBase, TomographySettings, DEFAULT_SOLVER_KWARGS
from grove.tomography import tomography
import grove.tomography.utils as ut
import grove.tomography.operator_utils as o_ut

_log = logging.getLogger(__name__)

qt = ut.import_qutip()
cvxpy = ut.import_cvxpy()


UNIT_TRACE = 'unit_trace'
POSITIVE = 'positive'
DEFAULT_STATE_TOMO_SETTINGS = TomographySettings(
    constraints={UNIT_TRACE},
    solver_kwargs=DEFAULT_SOLVER_KWARGS
)


def _prepare_c_jk_m(readout_povm, pauli_basis, channel_ops):
    """
    Prepare the coefficient matrix for state tomography. This function uses sparse matrices
    for much greater efficiency.
    The coefficient matrix is defined as:

    .. math::

            C_{(jk)m} = \tr{\Pi_{s_j} \Lambda_k(P_m)} = \sum_{r}\pi_{jr}(\mathcal{R}_{k})_{rm}

    where :math:`\Lambda_k(\cdot)` is the quantum map corresponding to the k-th pre-measurement
    channel, i.e., :math:`\Lambda_k(\rho) = E_k \rho E_k^\dagger` where :math:`E_k` is the k-th
    channel operator. This map can also be represented via its transfer matrix
    :math:`\mathcal{R}_{k}`. In that case one also requires the overlap between the (generalized)
    Pauli basis ops and the projection operators
    :math:`\pi_{jl}:=\sbraket{\Pi_j}{P_l} = \tr{\Pi_j P_l}`.
    See the grove documentation on tomography for detailed information.

    :param DiagonalPOVM readout_povm: The POVM corresponding to the readout plus classifier.
    :param OperatorBasis pauli_basis: The (generalized) Pauli basis employed in the estimation.
    :param list channel_ops: The pre-measurement channel operators as `qutip.Qobj`
    :return: The coefficient matrix necessary to set up the binomial state tomography problem.
    :rtype: scipy.sparse.csr_matrix
    """
    channel_transfer_matrices = [pauli_basis.transfer_matrix(qt.to_super(ek)) for ek in channel_ops]

    # This bit could be more efficient but does not run super long and is thus preserved for
    #  readability.
    pi_jr = csr_matrix(
        [pauli_basis.project_op(n_j).toarray().ravel()
         for n_j in readout_povm.ops])

    # Dict used for constructing our sparse matrix, keys are tuples (row_index, col_index), values
    # are the non-zero elements of the final matrix.
    c_jk_m_elms = {}

    # This explicitly exploits the sparsity of all operators involved
    for k in range(len(channel_ops)):
        pi_jr__rk_rm = (pi_jr * channel_transfer_matrices[k]).tocoo()
        for (j, m, val) in ut.izip(pi_jr__rk_rm.row, pi_jr__rk_rm.col, pi_jr__rk_rm.data):
            # The multi-index (j,k) is enumerated in column-major ordering (like Fortran arrays)
            c_jk_m_elms[(j + k * readout_povm.pi_basis.dim, m)] = val.real

    # create sparse matrix from COO-format (see scipy.sparse docs)
    _keys, _values = ut.izip(*c_jk_m_elms.items())
    _rows, _cols = ut.izip(*_keys)
    c_jk_m = coo_matrix((list(_values), (list(_rows), list(_cols))),
                        shape=(readout_povm.pi_basis.dim * len(channel_ops),
                               pauli_basis.dim)).tocsr()
    return c_jk_m


class StateTomography(TomographyBase):
    """
    A StateTomography object encapsulates the result of quantum state estimation from tomographic
    data. It provides convenience functions for visualization and computing state fidelities.
    """
    __tomography_type__ = "STATE"

    @staticmethod
    def estimate_from_ssr(histograms, readout_povm, channel_ops, settings):
        """
        Estimate a density matrix from single shot histograms obtained by measuring bitstrings in
        the Z-eigenbasis after application of given channel operators.

        :param numpy.ndarray histograms: The single shot histograms, `shape=(n_channels, dim)`.
        :param DiagognalPOVM readout_povm: The POVM corresponding to the readout plus classifier.
        :param list channel_ops: The tomography measurement channels as `qutip.Qobj`'s.
        :param TomographySettings settings: The solver and estimation settings.
        :return: The generated StateTomography object.
        :rtype: StateTomography
        """
        nqc = len(channel_ops[0].dims[0])
        pauli_basis = grove.tomography.operator_utils.PAULI_BASIS ** nqc
        pi_basis = readout_povm.pi_basis

        if not histograms.shape[1] == pi_basis.dim:  # pragma no coverage
            raise ValueError("Currently tomography is only implemented for two-level systems.")

        # prepare the log-likelihood function parameters, see documentation
        n_kj = np.asarray(histograms)
        c_jk_m = _prepare_c_jk_m(readout_povm, pauli_basis, channel_ops)
        rho_m = cvxpy.Variable(pauli_basis.dim)

        p_jk = c_jk_m * rho_m
        obj = -n_kj.ravel() * cvxpy.log(p_jk)

        p_jk_mat = cvxpy.reshape(p_jk, pi_basis.dim, len(channel_ops))  # cvxpy has col-major order

        # Default constraints:
        # MLE must describe valid probability distribution
        # i.e., for each k, p_jk must sum to one and be element-wise non-negative:
        # 1. \sum_j p_jk == 1  for all k
        # 2. p_jk >= 0         for all j, k
        # where p_jk = \sum_m c_jk_m rho_m
        constraints = [
            p_jk >= 0,
            np.matrix(np.ones((1, pi_basis.dim))) * p_jk_mat == 1,
        ]

        rho_m_real_imag = sum((rm * o_ut.to_realimag(Pm)
                               for (rm, Pm) in ut.izip(rho_m, pauli_basis.ops)), 0)

        if POSITIVE in settings.constraints:
            if tomography._SDP_SOLVER.is_functional():
                constraints.append(rho_m_real_imag >> 0)
            else:  # pragma no coverage
                _log.warning("No convex solver capable of semi-definite problems installed.\n"
                             "Dropping the positivity constraint on the density matrix.")

        if UNIT_TRACE in settings.constraints:
            # this assumes that the first element of the Pauli basis is always proportional to
            # the identity
            constraints.append(rho_m[0, 0] == 1. / pauli_basis.ops[0].tr().real)

        prob = cvxpy.Problem(cvxpy.Minimize(obj), constraints)

        _log.info("Starting convex solver")
        prob.solve(solver=tomography.SOLVER, **settings.solver_kwargs)
        if prob.status != cvxpy.OPTIMAL:  # pragma no coverage
            _log.warning("Problem did not converge to optimal solution. "
                         "Solver settings: {}".format(settings.solver_kwargs))

        return StateTomography(np.array(rho_m.value).ravel(), pauli_basis, settings)

    def __init__(self, rho_coeffs, pauli_basis, settings):
        """
        Construct a StateTomography to encapsulate the result of estimating the quantum state from
        a quantum tomography measurement.

        :param numpy.ndarray r_est: The estimated quantum state represented in a given (generalized)
        Pauli basis.
        :param OperatorBasis pauli_basis: The employed (generalized) Pauli basis.
        :param TomographySettings settings: The settings used to estimate the state.
        """
        self.rho_coeffs = rho_coeffs
        self.pauli_basis = pauli_basis
        self.rho_est = sum((r_m * p_m for r_m, p_m in ut.izip(rho_coeffs, pauli_basis.ops)))
        self.settings = settings

    def fidelity(self, other):
        """
        Compute the quantum state fidelity of the estimated state with another state.

        :param qutip.Qobj other: The other quantum state.
        :return: The fidelity, a real number between 0 and 1.
        :rtype: float
        """
        return qt.fidelity(self.rho_est, other)

    def plot_state_histogram(self, ax):
        """
        Visualize the complex matrix elements of the estimated state.

        :param matplotlib.Axes ax: A matplotlib Axes object to plot into.
        """
        title = "Estimated state"
        nqc = int(round(np.log2(self.rho_est.data.shape[0])))
        labels = ut.basis_labels(nqc)
        return ut.state_histogram(self.rho_est, ax, title)

    def plot(self):
        """
        Visualize the state.

        :return: The generated figure.
        :rtype: matplotlib.Figure
        """
        width = 10
        # The pleasing golden ratio.
        height = width / 1.618
        f = plt.figure(figsize=(width, height))
        ax = f.add_subplot(111, projection="3d")

        self.plot_state_histogram(ax)
        return f


def state_tomography_programs(state_prep, qubits=None,
                              rotation_generator=tomography.default_rotations):
    """
    Yield tomographic sequences that prepare a state with Quil program `state_prep` and then append
    tomographic rotations on the specified `qubits`. If `qubits is None`, it assumes all qubits in
    the program should be tomographically rotated.

    :param Program state_prep: The program to prepare the state to be tomographed.
    :param list|NoneType qubits: A list of Qubits or Numbers, to perform the tomography on. If
    `None`, performs it on all in state_prep.
    :param generator rotation_generator: A generator that yields tomography rotations to perform.
    :return: Program for state tomography.
    :rtype: Program
    """
    if qubits is None:
        qubits = state_prep.get_qubits()
    for tomography_program in rotation_generator(*qubits):
        state_tomography_program = Program(Pragma("PRESERVE_BLOCK"))
        state_tomography_program.inst(state_prep)
        state_tomography_program.inst(tomography_program)
        state_tomography_program.inst(Pragma("END_PRESERVE_BLOCK"))
        yield state_tomography_program


def do_state_tomography(preparation_program, nsamples, cxn, qubits=None, use_run=False):
    """
    Method to perform both a QPU and QVM state tomography, and use the latter as
    as reference to calculate the fidelity of the former.

    :param Program preparation_program: Program to execute.
    :param int nsamples: Number of samples to take for the program.
    :param QVMConnection|QPUConnection cxn: Connection on which to run the program.
    :param list qubits: List of qubits for the program.
    to use in the tomography analysis.
    :param bool use_run: If ``True``, use append measurements on all qubits and use ``cxn.run``
        instead of ``cxn.run_and_measure``.
    :return: The state tomogram.
    :rtype: StateTomography
    """
    return tomography._do_tomography(preparation_program, nsamples, cxn, qubits,
                                     tomography.MAX_QUBITS_STATE_TOMO,
                                     StateTomography, state_tomography_programs,
                                     DEFAULT_STATE_TOMO_SETTINGS, use_run=use_run)



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################


def isclose(a, b, rel_tol=1e-10, abs_tol=0.0):
    """
    Compares two parameter values.
    :param a: First parameter
    :param b: Second parameter
    :param rel_tol: Relative tolerance
    :param abs_tol: Absolute tolerance
    :return: Boolean telling whether or not the parameters are close enough to be the same
    """
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)



"""
This module helps construct sums of non-hermitian operators for measuring
specific sets of amplitudes.  The general procedure for measuring individual
amplitudes was outlined in

Nature Communications 7, Article number: 10439 (2016)
doi:10.1038/ncomms10439

Each amplitude c_{j} is associated with an operator |a><j| where |a> is an
arbitrary computation basis vector.  These are referred to as column operators.
(Cj)

Each column operator has an expectation value
<Cj> = <psi|a>c_{j}

which is proportional to ta complex state vector coefficient...except has
an extra factor <psi|a> that does not depend on 'j'.  Therefore, we can select
a single |a> and for C_{a} = |<psi|a>|^{2}.  What this means is that we can
figure out |<psi|a>|^{2} and thus |<psi|a>| which gives us the normalization
constant we want.
"""
from functools import reduce
import numpy as np
from pyquil.paulis import sX, sY, sZ, sI, PauliSum
from grove.measurements.estimation import estimate_locally_commuting_operator


def _single_projector_generator(ket_op, bra_op, index):
    """
    Generate the pauli sum terms corresponding to |ket_op><brak_op|

    :param ket_op: single qubit computational basis state
    :param bra_op: single qubit computational basis state
    :param index: qubit index to assign to the projector
    :return: pauli sum of single qubit projection operator
    :rtype: PauliSum
    """
    if not isinstance(ket_op, int):
        raise TypeError("ket_op needs to be an integer")
    if not isinstance(bra_op, int):
        raise TypeError("ket_op needs to be an integer")
    if ket_op not in [0, 1] or bra_op not in [0, 1]:
        raise ValueError("bra and ket op needs to be either 0 or 1")

    if ket_op == 0 and bra_op == 0:
        return 0.5 * (sZ(index) + sI(index))
    elif ket_op == 0 and bra_op == 1:
        return 0.5 * (sX(index) + 1j * sY(index))
    elif ket_op == 1 and bra_op == 0:
        return 0.5 * (sX(index) - 1j * sY(index))
    else:
        return 0.5 * (sI(index) - sZ(index))


def projector_generator(ket, bra):
    """
    Generate a Pauli Sum that corresponds to the projection operator |ket><bra|

    note: ket and bra are numerically ordered such that ket = [msd, ..., lsd]
          where msd == most significant digit and lsd = least significant digit.

    :param List ket: string of zeros and ones corresponding to a computational
                     basis state.
    :param List bra: string of zeros and ones corresponding to a computational
                     basis state.
    :return: projector as a pauli sum
    :rytpe: PauliSum
    """
    projectors = []
    for index, (ket_one_qubit, bra_one_qubit) in enumerate(zip(ket[::-1], bra[::-1])):
        projectors.append(_single_projector_generator(ket_one_qubit,
                                                      bra_one_qubit, index))
    return reduce(lambda x, y: x * y, projectors)


def measure_wf_coefficients(prep_program, coeff_list, reference_state,
                            quantum_resource, variance_bound=1.0E-6):
    """
    Measure a set of coefficients with a phase relative to the reference_state

    :param prep_program: pyQuil program to prepare the state
    :param coeff_list: list of integers labeling amplitudes to measure
    :param reference_state: Integer of the computational basis state to use as
                            a reference
    :param quantum_resource: An instance of a quantum abstract machine
    :param variance_bound: Default 1.0E-6.  variance of the monte carlo
                           estimator for the non-hermitian operator
    :return: returns a list of reference_state amplitude + coeff_list amplitudes
    """
    num_qubits = len(prep_program.get_qubits())
    normalizer_ops = projector_generator(reference_state, reference_state)
    c0_coeff, _, _ = estimate_locally_commuting_operator(
        prep_program, normalizer_ops, variance_bound=variance_bound,
        quantum_resource=quantum_resource)
    c0_coeff = np.sqrt(c0_coeff)

    amplitudes = []
    for ii in coeff_list:
        if ii == reference_state:
            amplitudes.append(c0_coeff)
        else:
            bra = list(map(int, np.binary_repr(ii, width=num_qubits)))
            c_ii_op = projector_generator(reference_state, bra)

            result = estimate_locally_commuting_operator(
                prep_program, c_ii_op, variance_bound=variance_bound,
                quantum_resource=quantum_resource)
            amplitudes.append(result[0] / c0_coeff)

    return amplitudes


def measure_pure_state(prep_program, reference_state, quantum_resource,
                       variance_bound=1.0E-6):
    """
    Measure the coefficients of the pure state

    :param prep_program: pyQuil program to prepare the state
    :param reference_state: Integer of the computational basis state to use as
                            a reference
    :param quantum_resource: An instance of a quantum abstract machine
    :param variance_bound: Default 1.0E-6.  variance of the monte carlo
                           estimator for the non-hermitian operator
    :return: an estimate of the wavefunction as a numpy.ndarray
    """
    num_qubits = len(prep_program.get_qubits())
    amplitudes_to_measure = list(range(2 ** num_qubits))

    amplitudes = measure_wf_coefficients(prep_program, amplitudes_to_measure,
                                         reference_state,
                                         quantum_resource,
                                         variance_bound=variance_bound)
    wavefunction = np.asarray(amplitudes)
    return wavefunction.reshape((-1, 1))
